from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from .action import Action

__all__ = ["Reducer"]


TAction = TypeVar("TAction", bound=Action)
"""Invariant type variable bound by an `Action`."""

TState = TypeVar("TState")
"""Invariant type variable for a generic state."""


class Reducer(Generic[TAction, TState], ABC):
    """A base class for a reducer.

    Reducer performs state mutations by applying an action to the current state.
    Defining a reducer is done by subclassing the `Reducer` and specyfing the type of
    action it should handle.

    Example:
        Following example defines a reducer for the `AddToCounterAction`, that adds
        the `action.value` to the `State.counter`.

        ```python
        from typing import NamedTuple

        import pydepot


        class State(NamedTuple):
            counter: int


        class AddToCounterAction(pydepot.Action):
            def __init__(self, value: int):
                self.value: int = value


        class AddToCounterReducer(pydepot.Reducer[AddToCounterAction, State]):
            @property
            def action_type(self) -> type[AddToCounterAction]:
                return AddToCounterAction

            def apply(self, action: AddToCounterAction, state: State) -> State:
                return State(counter=state.counter + action.value)  # The mutation.
        ```
    """

    __slots__ = ()

    @property
    @abstractmethod
    def action_type(self) -> type[TAction]:
        """The type of an action that this reducer handles."""

    @abstractmethod
    def apply(self, action: TAction, state: TState) -> TState:
        """Apply an action to the current state.

        The `Store` calls this method when an action is dispatched. The return value
        should be the state after applying the mutation described by the action.

        Args:
            action (TAction): The action to apply.
            state (TState): The current state.

        Returns:
            TState: The state after applying the action.
        """



from .abc.action import Action
from .abc.reducer import Reducer
from .abc.subscriber import StoreSubscriber
from .store import Store

__all__ = [
    "Action",
    "Reducer",
    "Store",
    "StoreSubscriber",
]



from typing import Any, Generic, TypeVar
from weakref import WeakSet

from .abc import Action, Reducer, StoreSubscriber

__all__ = ["Store"]


TAction = TypeVar("TAction", bound=Action)
"""Invariant type variable bound by an `Action`."""

TState = TypeVar("TState")
"""Invariant type variable for a generic state."""


class Store(Generic[TState]):
    """The centralized manager of the state.

    Notifies it's subscribers whenever the state changes.

    To perform a state mutation, call the `dispatch` method with an action.
    For an action to be applied a matching reducer has to be registered first.

    Tip:
        Make sure that the reducers are registered before dispatching actions.

    """

    __slots__ = ("_reducers", "_state", "_subscribers")

    @property
    def state(self) -> TState:
        """The current state."""
        return self._state

    def __init__(self, initial_state: TState):
        """Initialize new store with a initial state.

        Args:
            initial_state (TState): The initial state.
        """
        self._reducers: dict[type[Action], Reducer[Any, TState]] = {}
        self._state: TState = initial_state
        self._subscribers: WeakSet[StoreSubscriber[TState]] = WeakSet()

    def dispatch(self, action: Action) -> None:
        """Dispatch an action to the store.

        If the dispatched action causes the state to change, all current subscribers will
        get notified by the `on_state` method.

        Before dispatching the action, make sure a `Reducer` with the matching
        `action_type` has been registered.

        Args:
            action (Action): Action to dispatch.
        """
        if reducer := self._reducers.get(type(action), None):
            new_state = reducer.apply(action=action, state=self._state)
            if new_state != self._state:
                self._state = new_state
                for subscriber in self._subscribers.copy():
                    subscriber.on_state(self._state)

    def register(self, reducer: Reducer[TAction, TState]) -> None:
        """Register a reducer.

        When an action with a type equal to `reducer.action_type` is dispatched,
        the `apply` method will be called on the `reducer` to perform the mutation.

        Registering another reducer with the same `action_type` will override the
        previous one. The store keeps a strong reference to the `reducer`.

        Args:
            reducer (Reducer[TAction, TState]): The reducer to register.
        """
        self._reducers[reducer.action_type] = reducer

    def subscribe(
        self, subscriber: StoreSubscriber[TState], include_current: bool = False
    ) -> None:
        """Subscribe to state updates.

        When the state changes, the `subscriber.on_state` will be called with the
        new state. If `include_current` is true, the subscriber is notified immediately
        with the current state.

        This method is idempotent - multiple calls will not cause multiple subscriptions.

        Make sure to keep a strong reference to the subscriber, because the store only
        keeps a weak reference.

        The `StoreSubscriber` is structurally typed, so any object implementing
        the `on_state(state: TState) -> None` method is a valid subscriber.

        Args:
            subscriber (StoreSubscriber[TState]): The subscriber to subscribe.
            include_current (bool): If true, the subscriber is immediately notified with
                the current state. If false, the first notification will happen on the
                next state change.
        """
        self._subscribers.add(subscriber)
        if include_current:
            subscriber.on_state(self._state)

    def unsubscribe(self, subscriber: StoreSubscriber[TState]) -> None:
        """Unsubscribe from state updates.

        The subscriber will no longer be notified when the state changes.

        Args:
            subscriber (StoreSubscriber[TState]): The subscriber to unsubscribe.
        """
        self._subscribers.discard(subscriber)



from collections.abc import Callable
from typing import NamedTuple
from unittest.mock import MagicMock, PropertyMock, create_autospec

import pytest

from src.pydepot import Action, Reducer, Store, StoreSubscriber
from tests.helpers import not_raises


class MockState(NamedTuple):
    value: int


class FooAction(Action):
    @property
    def value(self) -> int:
        return 42


class BarAction(Action):
    @property
    def value(self) -> int:
        raise NotImplementedError


@pytest.fixture()
def create_reducer() -> Callable[[type[Action] | None], Reducer[Action, MockState]]:
    def factory(
        action_type: type[Action] | None = None,
    ) -> Reducer[Action, MockState]:
        def apply(action, state):
            return MockState(value=action.value)

        reducer = create_autospec(Reducer)
        type(reducer).action_type = PropertyMock(return_value=action_type or FooAction)
        reducer.apply = MagicMock(side_effect=apply)

        return reducer

    return factory


@pytest.fixture()
def create_sut() -> Callable[[MockState | None], Store]:
    def factory(state: MockState | None = None) -> Store:
        return Store(initial_state=state or MockState(value=0))

    return factory


@pytest.fixture()
def create_subsciber() -> Callable[[], StoreSubscriber]:
    return lambda: create_autospec(StoreSubscriber)


class TestDispatch:
    def test_when_no_registered_reducers__does_not_raise(self, create_sut) -> None:
        sut = create_sut()
        with not_raises(Exception):
            sut.dispatch(FooAction())

    def test_when_reducer_registered_for_different_action__does_not_call_apply(
        self, create_sut, create_reducer
    ) -> None:
        reducer = create_reducer(action_type=BarAction)
        sut = create_sut()
        sut.register(reducer)

        sut.dispatch(FooAction())

        reducer.apply.assert_not_called()

    def test_when_reducer_registered_for_action__calls_apply(
        self, create_sut, create_reducer
    ) -> None:
        action = FooAction()
        state = MockState(value=0)
        reducer = create_reducer(action_type=FooAction)
        sut = create_sut(state=state)
        sut.register(reducer)

        sut.dispatch(action)

        reducer.apply.assert_called_with(action=action, state=state)

    def test_when_reducer_registered_for_action__new_state_equals_current__does_not_notify_subscribers(  # noqa: E501
        self, create_sut, create_subsciber, create_reducer
    ) -> None:
        subscriber = create_subsciber()
        reducer = create_reducer(action_type=FooAction)
        sut = create_sut(MockState(value=42))
        sut.subscribe(subscriber)
        sut.register(reducer)

        sut.dispatch(FooAction())

        subscriber.on_state.assert_not_called()

    def test_when_reducer_registered_for_action__new_state_not_equals_current__notifies_all_subscribers(  # noqa: E501
        self, create_sut, create_subsciber, create_reducer
    ) -> None:
        subscriber_1 = create_subsciber()
        subscriber_2 = create_subsciber()
        reducer = create_reducer(action_type=FooAction)
        sut = create_sut(MockState(value=0))
        sut.subscribe(subscriber_1)
        sut.subscribe(subscriber_2)
        sut.register(reducer)
        expected_new_state = MockState(value=42)

        sut.dispatch(FooAction())

        subscriber_1.on_state.assert_called_once_with(state=expected_new_state)
        subscriber_2.on_state.assert_called_once_with(state=expected_new_state)


class TestRegisterReducer:
    def test_when_action_dispatched_after_registering__apply_is_called(
        self, create_sut, create_reducer
    ) -> None:
        reducer = create_reducer(action_type=FooAction)
        sut = create_sut()

        sut.register(reducer)
        sut.dispatch(FooAction())

        reducer.apply.assert_called_once()

    def test_when_registered_same_reducer_twice__dispatching_actions_call_apply_once(
        self, create_sut, create_reducer
    ) -> None:
        reducer = create_reducer(action_type=FooAction)
        sut = create_sut()

        sut.register(reducer)
        sut.register(reducer)
        sut.dispatch(FooAction())

        reducer.apply.assert_called_once()

    def test_when_registered_two_reducers_for_same_action__dispatching_actions_call_apply_on_last_registered(  # noqa: E501
        self, create_sut, create_reducer
    ) -> None:
        reducer_1 = create_reducer(action_type=FooAction)
        reducer_2 = create_reducer(action_type=FooAction)
        sut = create_sut()

        sut.register(reducer_1)
        sut.register(reducer_2)
        sut.dispatch(FooAction())

        reducer_1.apply.assert_not_called()
        reducer_2.apply.assert_called_once()


class TestSubscribe:
    def test_when_include_current_false__does_not_notify(
        self, create_sut, create_subsciber
    ) -> None:
        sut = create_sut()
        subscriber = create_subsciber()

        sut.subscribe(subscriber, include_current=False)

        subscriber.on_state.assert_not_called()

    def test_when_include_current_true__notifies_with_current_state(
        self, create_sut, create_subsciber
    ) -> None:
        sut = create_sut()
        subscriber = create_subsciber()

        sut.subscribe(subscriber, include_current=True)

        subscriber.on_state.assert_called_once_with(state=sut.state)

    def test_when_subscribing_twice__dispatching_actions_notify_once(
        self, create_sut, create_subsciber, create_reducer
    ) -> None:
        reducer = create_reducer(action_type=FooAction)
        sut = create_sut()
        sut.register(reducer)
        subscriber = create_subsciber()

        sut.subscribe(subscriber, include_current=False)
        sut.subscribe(subscriber, include_current=False)
        sut.dispatch(FooAction())

        subscriber.on_state.assert_called_once()


class TestUnsubscribe:
    def test_when_unsubscribing_not_previously_subscribed__does_not_raise(
        self, create_sut, create_subsciber
    ) -> None:
        sut = create_sut()
        subscriber = create_subsciber()
        with not_raises(Exception):
            sut.unsubscribe(subscriber)

    def test_when_action_dispatched_after_unsubscribing__does_not_notify(
        self, create_sut, create_subsciber, create_reducer
    ) -> None:
        reducer = create_reducer(action_type=FooAction)
        subscriber = create_subsciber()
        sut = create_sut()
        sut.register(reducer)
        sut.subscribe(subscriber, include_current=False)

        sut.unsubscribe(subscriber)
        sut.dispatch(FooAction())

        subscriber.on_state.assert_not_called()

    def test_when_action_dispatched_after_unsubscribing_twice__does_not_notify(
        self, create_sut, create_subsciber, create_reducer
    ) -> None:
        reducer = create_reducer(action_type=FooAction)
        subscriber = create_subsciber()
        sut = create_sut()
        sut.register(reducer)
        sut.subscribe(subscriber, include_current=False)

        sut.unsubscribe(subscriber)
        sut.unsubscribe(subscriber)
        sut.dispatch(FooAction())

        subscriber.on_state.assert_not_called()



from abc import ABC

__all__ = ["Action"]


class Action(ABC):
    """A base class for an action.

    An action represents an intent to perfrom a state mutation.
    Defining an action is done by subclassing the `Action` with a name of the mutation.

    Example:
        Following example defines an action that, when applied, should add a `value` to
        the `counter` property of some state.
        ```python
        import pydepot

        class AddToCounterAction(pydepot.Action):
            def __init__(self, value: int):
                self.value: int = value
        ```
    """

    __slots__ = ()



import pathlib
import nox

SOURCE_DIR = "src/pydepot"
PYTHON_DEFAULT_VERSION = "3.11"
LINT_DEPENDENCIES = [
    "black==23.7.0",
    "mypy==1.5.0",
    "ruff==0.0.284",
]
RESULTS_DIR = pathlib.Path('./reports').resolve()
RESULTS_DIR.mkdir(exist_ok=True)

@nox.session(python=PYTHON_DEFAULT_VERSION)
def test(session: nox.Session) -> None:
    session.install("pytest", "pytest-cov")
    output_file = RESULTS_DIR / "test.xml"
    session.run("pytest", "--junit-xml", str(output_file), "--cov=src.pydepot", "--cov-report=term-missing")
    session.notify("coverage")


@nox.session(python=PYTHON_DEFAULT_VERSION)
def coverage(session: nox.Session) -> None:
    session.install("coverage[toml]")
    output_file = RESULTS_DIR / "coverage.xml"
    session.run("coverage", "xml", "-o", str(output_file))
    session.run("coverage", "erase")


@nox.session(python=PYTHON_DEFAULT_VERSION)
def lint(session: nox.Session) -> None:
    session.install(*LINT_DEPENDENCIES)
    lint_output = RESULTS_DIR / "lint.xml"
    session.run("ruff", "check", "--format", "junit", "--output-file", str(lint_output), "src")

    session.run("black", "--check", "tests", "src/pydepot")

    typing_output = RESULTS_DIR / "typing.xml"
    session.run("mypy", "--junit-xml", str(typing_output), SOURCE_DIR)



from abc import abstractmethod
from typing import Protocol, TypeVar

__all__ = ["StoreSubscriber"]


TState = TypeVar("TState", contravariant=True)
"""Contravariant type variable for a generic state."""


class StoreSubscriber(Protocol[TState]):
    """The protocol for a store subscriber.

    Any object implementing the `on_state(state: TState) -> None` method
    is a valid subscriber.
    """

    @abstractmethod
    def on_state(self, state: TState) -> None:
        """Notify the subscriber of a state change.

        If the subscriber is currently subscribed to a `Store`, this method is called
        every time the state changes.

        Args:
            state (TState): The new state.
        """



from contextlib import contextmanager

import pytest


@contextmanager
def not_raises(exception):
    try:
        yield
    except exception:
        raise pytest.fail(f"Did raise {exception!r}")  # noqa: TRY200,TRY003
